@@ -17,7 +17,7 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
+;; along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -147,11 +147,6 @@ Please manually run the hook `exwm-workspace-list-change-hook' afterwards.")
 
 (defvar exwm-workspace--workareas nil "Workareas (struts excluded).")
 
-(defvar exwm-workspace--frame-y-offset 0
-  "Offset between Emacs inner & outer frame in Y.")
-(defvar exwm-workspace--window-y-offset 0
-  "Offset between Emacs first window & outer frame in Y.")
-
 (defvar exwm-input--event-hook)
 (defvar exwm-layout-show-all-buffers)
 (defvar exwm-manage--desktop)
@@ -425,34 +420,6 @@ Show PROMPT to the user if non-nil."
   (exwm--log "%s" exwm-workspace--workareas)
   (run-hooks 'exwm-workspace--update-workareas-hook))
 
-(defun exwm-workspace--update-offsets ()
-  "Update `exwm-workspace--frame-y-offset'/`exwm-workspace--window-y-offset'."
-  (exwm--log)
-  (if (not (and exwm-workspace--list
-                (or menu-bar-mode tool-bar-mode)))
-      (setq exwm-workspace--frame-y-offset 0
-            exwm-workspace--window-y-offset 0)
-    (redisplay t)
-    (let* ((frame (elt exwm-workspace--list 0))
-           (edges (window-inside-absolute-pixel-edges (frame-first-window
-                                                       frame))))
-      (with-slots (y)
-          (xcb:+request-unchecked+reply exwm--connection
-              (make-instance 'xcb:GetGeometry
-                             :drawable (frame-parameter frame
-                                                        'exwm-container)))
-        (with-slots ((y* y))
-            (xcb:+request-unchecked+reply exwm--connection
-                (make-instance 'xcb:GetGeometry
-                               :drawable (frame-parameter frame
-                                                          'exwm-outer-id)))
-          (with-slots ((y** y))
-              (xcb:+request-unchecked+reply exwm--connection
-                  (make-instance 'xcb:GetGeometry
-                                 :drawable (frame-parameter frame 'exwm-id)))
-            (setq exwm-workspace--frame-y-offset (- y** y*)
-                  exwm-workspace--window-y-offset (- (elt edges 1) y))))))))
-
 (defun exwm-workspace--set-active (frame active)
   "Make frame FRAME active on its monitor.
 ACTIVE indicates whether to set the frame active or inactive."
@@ -725,11 +692,11 @@ Passing a workspace frame as the first option is for internal use only."
   (if (or (framep frame-or-index)
           (< frame-or-index (exwm-workspace--count)))
       (exwm-workspace-switch frame-or-index)
-    (let ((exwm-workspace--create-silently t))
-      (dotimes (_ (min exwm-workspace-switch-create-limit
-                       (1+ (- frame-or-index
-                              (exwm-workspace--count)))))
-        (make-frame))
+    (let ((count (1+ (- frame-or-index (exwm-workspace--count))))
+          (exwm-workspace--create-silently t))
+      (when (< exwm-workspace-switch-create-limit count)
+        (user-error "Cannot implicitly create more than %d workspaces at a time, customize `exwm-workspace-switch-create-limit' to change this limit" exwm-workspace-switch-create-limit))
+      (dotimes (_ count) (make-frame))
       (run-hooks 'exwm-workspace-list-change-hook))
     (exwm-workspace-switch frame-or-index)))
 
@@ -1410,12 +1377,21 @@ ALIST is an action alist, as accepted by function `display-buffer'."
 
 (defun exwm-workspace--get-next-workspace (frame)
   "Return the next workspace if workspace FRAME were removed.
-Return nil if FRAME is the only workspace."
+Return nil there are no other worksapces and/or all other workspaces
+are currently visible on other monitors."
   (let* ((index (exwm-workspace--position frame))
-         (lastp (= index (1- (exwm-workspace--count))))
-         (nextw (elt exwm-workspace--list (+ index (if lastp -1 +1)))))
-    (unless (eq frame nextw)
-      nextw)))
+         (count (exwm-workspace--count)))
+    (or
+     (cl-loop for i from (1+ index) below count
+              for nextw = (elt exwm-workspace--list i)
+              unless (or (eq frame nextw)
+                         (exwm-workspace--active-p nextw))
+              return nextw)
+     (cl-loop for i from (1- index) downto 0
+              for nextw = (elt exwm-workspace--list i)
+              unless (or (eq frame nextw)
+                         (exwm-workspace--active-p nextw))
+              return nextw))))
 
 (defun exwm-workspace--remove-frame-as-workspace (frame &optional quit)
   "Stop treating FRAME as a workspace.
@@ -1444,8 +1420,11 @@ When QUIT is non-nil cleanup avoid communicating with the X server."
           (when (memq other-frame following-frames)
             (exwm-workspace--set-desktop (car pair)))))
       ;; If the current workspace is deleted, switch to next one.
-      (when (eq frame exwm-workspace--current)
-        (exwm-workspace-switch next-frame))))
+      (if (eq frame exwm-workspace--current)
+          (exwm-workspace-switch next-frame)
+        ;; Otherwise, update the workspace index.
+        (setq exwm-workspace-current-index
+              (exwm-workspace--position next-frame)))))
   ;; Reparent out the frame.
   (let ((outer-id (frame-parameter frame 'exwm-outer-id)))
     (xcb:+request exwm--connection
@@ -1495,7 +1474,8 @@ When QUIT is non-nil cleanup avoid communicating with the X server."
   "Make workspace FRAME fullscreen.
 Called from a timer."
   (when (frame-live-p frame)
-    (set-frame-parameter frame 'fullscreen 'fullboth)))
+    (set-frame-parameter frame 'fullscreen 'fullboth)
+    (exwm-workspace--set-fullscreen frame)))
 
 (defun exwm-workspace--on-after-make-frame (frame)
   "Hook run upon `make-frame' that configures FRAME as a workspace."
